import itertools as it
from functools import reduce
from typing import Any, Protocol, SupportsInt

import matplotlib.pyplot as plt
import numpy as np
from mpltools import annotation
from numpy.typing import NDArray

rk4a = [
    0.0,
    -567301805773.0 / 1357537059087.0,
    -2404267990393.0 / 2016746695238.0,
    -3550918686646.0 / 2091501179385.0,
    -1275806237668.0 / 842570457699.0,
]

rk4b = [
    1432997174477.0 / 9575080441755.0,
    5161836677717.0 / 13612068292357.0,
    1720146321549.0 / 2090206949498.0,
    3134564353537.0 / 4481467310338.0,
    2277821191437.0 / 14882151754819.0,
]

rk4c = [
    0.0,
    1432997174477.0 / 9575080441755.0,
    2526269341429.0 / 6820363962896.0,
    2006345519317.0 / 3224310063776.0,
    2802321613138.0 / 2924317926251.0,
]

pa = {"fill": False, "edgecolor": "black"}
ta = {"fontsize": 10}


def mul(*args: NDArray[Any]) -> NDArray[Any]:
    return reduce(np.matmul, args)


def init() -> None:
    np.set_printoptions(suppress=True)


class SolverMain(Protocol):
    """Protocol for solver main functions."""

    def __call__(self, p: SupportsInt, ne: SupportsInt) -> float: ...


def run_solver(
    main_fn: SolverMain,
    output_file: str,
    p_start: int = 1,
    p_end: int = 4,
    n_power_start: int = 2,
    n_power_end: int = 7,
) -> None:
    """Run a convergence study for a DG solver and plot the results.

    Args:
        main_fn: The solver's main function that takes (p, ne) and returns error.
        output_file: Filename for the output PDF plot.
        p_start: Starting polynomial order (inclusive).
        p_end: Ending polynomial order (exclusive).
        n_power_start: Starting power of 2 for number of elements (inclusive).
        n_power_end: Ending power of 2 for number of elements (exclusive).
    """
    P = np.arange(p_start, p_end, 1)
    N = 2 ** np.arange(n_power_start, n_power_end)
    E = np.zeros((len(P), len(N)))

    for (i, p), (j, n) in it.product(enumerate(P), enumerate(N)):
        E[i, j] = main_fn(p, n)

    _fig = plt.figure()
    ax = plt.gca()
    colors = ["b", "g", "r", "c", "m", "y", "k", "w"]

    for j, p in enumerate(P):
        plt.loglog(1.0 / N, E[j, :], label=f"p={p}", color=colors[p])
        slope, _intercept = np.polyfit(np.log(1.0 / N[-3:]), np.log(E[j, -3:]), 1)
        annotation.slope_marker(
            (1.0 / N[-1], E[j, -1]),
            (f"{slope:.2f}", 1),
            ax=ax,
            poly_kwargs=pa,
            text_kwargs=ta,
        )

    plt.tight_layout()
    plt.autoscale()
    plt.legend(loc="lower right")
    plt.savefig(output_file)



"""
Integration tests for DG solver main() functions.

These tests verify that each solver produces the expected error
for a fixed polynomial order (p) and number of elements (ne).
The expected errors are computed from reference runs.
"""

import pytest

from udg import entropydg, nodaldg, semdg, upwinddg

# Fixed test parameters
P = 2  # polynomial order
NE = 8  # number of elements

# Expected errors (computed from reference runs with p=2, ne=8)
# These are regression values - if the algorithm changes, update these
EXPECTED_ERRORS = {
    "semdg": 0.0060251910474230464,
    "nodaldg": 0.0017987486331684567,
    "upwinddg": 0.025562225115547704,
    "entropydg": 0.006025191047422863,
}

# Tolerance for floating-point comparison
RTOL = 1e-10


class TestSolverIntegration:
    """Integration tests for solver main() functions."""

    def test_semdg_error(self):
        """Test DGSEM solver produces expected error."""
        error = semdg.main(P, NE)
        assert error == pytest.approx(EXPECTED_ERRORS["semdg"], rel=RTOL)

    def test_nodaldg_error(self):
        """Test Nodal-DG solver produces expected error."""
        error = nodaldg.main(P, NE)
        assert error == pytest.approx(EXPECTED_ERRORS["nodaldg"], rel=RTOL)

    def test_upwinddg_error(self):
        """Test Upwind-DG solver produces expected error."""
        error = upwinddg.main(P, NE)
        assert error == pytest.approx(EXPECTED_ERRORS["upwinddg"], rel=RTOL)

    def test_entropydg_error(self):
        """Test Entropy-DG solver produces expected error."""
        error = entropydg.main(P, NE)
        assert error == pytest.approx(EXPECTED_ERRORS["entropydg"], rel=RTOL)


class TestSolverConvergence:
    """Test that solvers exhibit expected convergence behavior."""

    @pytest.mark.parametrize(
        "solver_main", [semdg.main, nodaldg.main, upwinddg.main, entropydg.main]
    )
    def test_error_decreases_with_refinement(self, solver_main):
        """Verify error decreases when mesh is refined (h-refinement)."""
        error_coarse = solver_main(P, 4)
        error_fine = solver_main(P, 8)
        assert error_fine < error_coarse, "Error should decrease with mesh refinement"

    @pytest.mark.parametrize(
        "solver_main", [semdg.main, nodaldg.main, upwinddg.main, entropydg.main]
    )
    def test_error_decreases_with_higher_order(self, solver_main):
        """Verify error decreases with higher polynomial order (p-refinement)."""
        error_low_order = solver_main(1, NE)
        error_high_order = solver_main(3, NE)
        assert error_high_order < error_low_order, (
            "Error should decrease with higher polynomial order"
        )



"""
Nodal-DG method
"""

from typing import SupportsInt

import numpy as np

from udg.common import mul, rk4a, rk4b, rk4c, run_solver
from udg.quadratures import jac_nodal_basis_at, nodal_basis_at, ortho_basis_at, zwglj


def main(_p: SupportsInt, _ne: SupportsInt) -> float:
    # number of coefficients inside each element
    p, ne = int(_p), int(_ne)
    K = p + 1

    # number of quadrature points inside each element
    Nq = K

    # define 1D mesh (We just need a sorted point distribution)
    xmesh = np.linspace(0, 1, ne + 1)

    # length of the domain
    _xlo, _xhi = np.min(xmesh), np.max(xmesh)

    # number of elements
    E = len(xmesh) - 1

    # left/right face maps
    mapL, mapR = np.arange(E + 1) * Nq - 1, np.arange(E + 1) * Nq
    mapL[0], mapR[-1] = 0, E * Nq - 1

    # size of each element
    h = np.diff(xmesh).reshape(-1, 1)
    assert np.min(h) > 1e-8, "Too small element size"

    # jacobian of the mapping from D^{st}=[-1,1] to D
    jac, invjac = h / 2, 2 / h

    # the quadrature (points, weights) on standard interval
    z, w = zwglj(Nq, 0.0, 0.0)

    # define the orthonormal basis
    B = nodal_basis_at(K, z, z)

    # define the mass matrix (see Karniadakis, page 122)
    # since the basis is orthonormal, this should be identity
    M = np.einsum("mq,q,lq->ml", B, w, B)
    invM = np.linalg.inv(M)

    # define the {K-1} order derivative matrix
    Dr = jac_nodal_basis_at(K, z, z)[0]

    # the nodal-dg due to Hasthaven
    vdm = ortho_basis_at(K, z).T
    invM = np.einsum("jn,in->ij", vdm, vdm)
    M = np.linalg.inv(invM)
    gLD, gRD = invM[0, :], invM[-1, :]

    # advection velocity
    advx = 1

    # initial time, time step, final time
    ti, dt, tf = 0.0, 5e-4, 1
    nsteps = int(np.ceil((tf - ti) / dt))
    dt = (tf - ti) / nsteps

    # Compute the initial solution (ucoeff and usol are same)
    xsol = np.array([0.5 * (xmesh[j] + xmesh[j + 1]) + jac[j] * z for j in range(E)])
    usol = np.sin(2 * np.pi * (xsol - 0))

    # Actual algorithm
    time = ti

    uprev = np.zeros_like(usol)

    for _step in range(nsteps):
        for istp in range(5):
            _tlocal = time + rk4c[istp] * dt

            # Step:1 extract the solution at interfaces
            uL = usol.reshape(-1, 1)[mapL]
            uR = usol.reshape(-1, 1)[mapR]

            # Insert the boundary condition information at left boundary
            uL[0], uR[-1] = uR[-1], uL[0]

            # Compute the upwind flux at interfaces
            fL, fR = advx * uL, advx * uR
            fupw = 0.5 * (fL + fR) + 0.5 * np.abs(advx) * (uL - uR)

            # Step:2 evaluate the derivative
            fx = advx * np.matmul(usol, Dr)

            # Step:3 Compute the jumps
            jR = fupw - fL
            jL = fupw - fR

            # Compute the continuous flux for each element
            # See the strong form eq 2.6 in (Hasthaven 2008)
            for e in range(E):
                fx[e, :] += jR[e + 1] * gRD - jL[e] * gLD
            fx *= -invjac

            uprev = rk4a[istp] * uprev + dt * fx
            usol += rk4b[istp] * uprev

        # increment time
        time += dt

    # exact solution
    usol_e = np.sin(2 * np.pi * (xsol - tf))
    error = np.dot(mul(np.fabs(usol - usol_e), w), jac)
    return float(error[0])


def run() -> None:
    run_solver(main, "nodaldg_adv.pdf")



"""
CLI entry points for UDG solvers.
"""

import click

from udg import entropydg, nodaldg, semdg, upwinddg


@click.group()
def cli() -> None:
    """UDG - Upwind Discontinuous Galerkin Method solvers."""


@cli.command("sem-dg")
def sem_dg_command() -> None:
    """Run the DGSEM solver (Kopriva et al., 2009)."""
    semdg.run()


@cli.command("nodal-dg")
def nodal_dg_command() -> None:
    """Run the Nodal-DG solver (Hesthaven et al., 2008)."""
    nodaldg.run()


@cli.command("upwind-dg")
def upwind_dg_command() -> None:
    """Run the Upwind-DG solver."""
    upwinddg.run()


@cli.command("entropy-dg")
def entropy_dg_command() -> None:
    """Run the Entropy-DG solver (Gassner et al., 2011)."""
    entropydg.run()


if __name__ == "__main__":
    cli()



# mypy: ignore-errors
from collections.abc import Iterable
from math import cos, gamma, pi

import numpy as np


def jacobi(n, a, b, z):
    j = [1]

    if n >= 1:
        j.append(((a + b + 2) * z + a - b) / 2)
    if n >= 2:
        apb, bbmaa = a + b, b * b - a * a

        for q in range(2, n + 1):
            qapbpq, apbp2q = q * (apb + q), apb + 2 * q
            apbp2qm1, apbp2qm2 = apbp2q - 1, apbp2q - 2

            aq = apbp2q * apbp2qm1 / (2 * qapbpq)
            bq = apbp2qm1 * bbmaa / (2 * qapbpq * apbp2qm2)
            cq = apbp2q * (a + q - 1) * (b + q - 1) / (qapbpq * apbp2qm2)

            # Update
            j.append((aq * z - bq) * j[-1] - cq * j[-2])

    return j


def jacobi_diff(n, a, b, z):
    dj = [0]

    if n >= 1:
        dj.extend(jp * (i + a + b + 2) / 2 for i, jp in enumerate(jacobi(n - 1, a + 1, b + 1, z)))

    return dj


def ortho_basis_at_py(order, p):
    jp = jacobi(order - 1, 0, 0, p)
    return [np.sqrt(i + 0.5) * p for i, p in enumerate(jp)]


def jac_ortho_basis_at_py(order, p):
    djp = jacobi_diff(order - 1, 0, 0, p)
    return [(np.sqrt(i + 0.5) * p,) for i, p in enumerate(djp)]


def ortho_basis_at(order, pts):
    if len(pts) and not isinstance(pts[0], Iterable):
        pts = [(p,) for p in pts]
        return np.array([ortho_basis_at_py(order, *p) for p in pts]).T
    return None


def jac_ortho_basis_at(order, pts):
    if len(pts) and not isinstance(pts[0], Iterable):
        pts = [(p,) for p in pts]
    J = [jac_ortho_basis_at_py(order, *p) for p in pts]
    return np.array(J).swapaxes(0, 2)


def nodal_basis_at(order, pts, epts):
    return np.linalg.solve(ortho_basis_at(order, pts), ortho_basis_at(order, epts)).T


def jac_nodal_basis_at(order, pts, epts):
    return np.linalg.solve(ortho_basis_at(order, pts), jac_ortho_basis_at(order, epts))


def jacobz(n, a, b):
    if not n:
        return []

    z = np.zeros(n)
    dth = pi / (2.0 * n)
    rlast = 0.0

    for k in range(n):
        r = -cos((2.0 * k + 1.0) * dth)
        if k >= 1:
            r = 0.5 * (r + rlast)

        for _j in range(1, 5000):
            poly = jacobi(n, a, b, r)[-1]
            pder = jacobi_diff(n, a, b, r)[-1]
            tsum = 0.0
            for i in range(k):
                tsum += 1.0 / (r - z[i])

            delr = -poly / (pder - tsum * poly)
            r += delr
            if abs(delr) < 1e-20:
                break

        z[k] = r
        rlast = r
    return z


"""Computation of gauss quadratures via eigenvalue decomposition.
Ref: Orthogonal Polynomials: Computation and Approximation, Walter Gautschi"""


def rjacobi(n, a, b):
    ra, rb = np.zeros(n), np.zeros(n)

    apbp2 = 2.0 + a + b
    ra[0] = (b - a) / apbp2
    rb[0] = np.power(2.0, a + b + 1) * (gamma(a + 1.0) * gamma(b + 1.0) / gamma(apbp2))
    rb[1] = 4.0 * (a + 1.0) * (b + 1.0) / ((apbp2 + 1.0) * apbp2 * apbp2)
    # Compute other terms
    apbp2 += 2
    for i in range(1, n - 1):
        ra[i] = (b * b - a * a) / ((apbp2 - 2.0) * apbp2)
        rb[i + 1] = (
            4.0
            * (i + 1)
            * (i + 1 + a)
            * (i + 1 + b)
            * (i + 1 + a + b)
            / ((apbp2 * apbp2 - 1) * apbp2 * apbp2)
        )
        apbp2 += 2

    ra[n - 1] = (b * b - a * a) / ((apbp2 - 2.0) * apbp2)
    return ra, rb


def gauss(_n, ra, rb):
    scal = rb[0]
    rb[:-1] = np.sqrt(rb[1:])
    z, V = np.linalg.eigh(np.diag(ra) + np.diag(rb[:-1], -1))
    zidx = np.argsort(z)
    z.sort()
    V = V[:, zidx]
    w = V[0, :]
    w = scal * (w**2)
    return z, w


def zwgj(n, a, b):
    ra, rb = rjacobi(n, a, b)
    z, w = gauss(n, ra, rb)
    return z, w


def zwglj(n, a, b):
    N = n - 2
    z, w = rjacobi(n, a, b)
    apb1 = a + b + 1.0
    z[n - 1] = (a - b) / (2.0 * N + apb1 + 1.0)
    w[n - 1] = (
        4.0
        * (N + a + 1.0)
        * (N + b + 1.0)
        * (N + apb1)
        / ((2.0 * N + apb1) * np.power(2 * N + apb1 + 1, 2.0))
    )
    z, w = gauss(n, z, w)
    return z, w
